/************************************************************
* This file is part of the XDN Router project.
*
* XDN Router is a packet routing device to be used with
* a local XDN (XBee Distributed Network) configuration.
* It supports both core and edge routing modes.
* Copyright (C) 2008 Adam Shapiro, Cornell University
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
************************************************************/
/** \file
 * \brief Link layer implementation.
 *
 * This file contains implementation code for
 * the XDN Link layer.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdint.h>
#include "link.h"
#include "router.h"
#include "link_defs.h"

volatile uint16_t link_rxHead[LINK_NUM_INTERFACES],
    link_rxTail[LINK_NUM_INTERFACES],
    link_rxInsertionPoint[LINK_NUM_INTERFACES],
    link_rxBytesPending[LINK_NUM_INTERFACES];
volatile uint8_t link_rxLength[LINK_NUM_INTERFACES];
volatile uint8_t link_rxQueue[LINK_NUM_INTERFACES][LINK_RX_QUEUE_SIZE];
uint8_t link_rxPayloadBuffer[LINK_MTU_SIZE];
volatile uint8_t link_rxChecksum[LINK_NUM_INTERFACES];

volatile uint8_t link_activityTimer[LINK_NUM_INTERFACES];
int8_t link_moduleStatus[LINK_NUM_INTERFACES];

volatile uint16_t link_txHead[LINK_NUM_INTERFACES],
    link_txTail[LINK_NUM_INTERFACES],
    link_txRemaining[LINK_NUM_INTERFACES];
volatile uint8_t link_txQueue[LINK_NUM_INTERFACES][LINK_TX_QUEUE_SIZE];
uint8_t link_txPayloadBuffer[LINK_MTU_SIZE];

PendingATResponse pendingCmd;

void Link_Init(uint8_t interface)
{
    uint8_t data[6];

    data[0]=XBEE_API_AT_COMMAND;//API identifier.
    data[1]=0;//Frame ID (0 to disable response).

    //Initialize RX queue.
    link_rxHead[interface]=0;
    link_rxTail[interface]=0;
    link_rxInsertionPoint[interface]=0;
    link_rxLength[interface]=0;
    link_rxBytesPending[interface]=0;
    link_activityTimer[interface]=0;
    link_moduleStatus[interface]=-1;

    //Initialize TX queue.
    link_txHead[interface]=0;
    link_txTail[interface]=0;
    link_txRemaining[interface]=LINK_TX_QUEUE_SIZE;

//copy debug

#if DEVICE_TYPE==0

		//Bring up host interface.
	    DEBUG_ON(0);

        //Initialize to specified baud rate, 8N1.
        UBRR0=BAUD(LINK_BAUD_RATE);
        
        //Enable RX and TX with RX interrupt only.
        //The TX (DRE) interrupt is enabled when sending data.
        UCSR0B=_BV(RXEN0) | _BV(TXEN0) | _BV(RXCIE0);

        //Reset module to clear previous settings.
        Link_ResetModule(0);

        //Module local address. Must be set to 0xFFFF
        //to enable 64-bit addressing.
        XBEE_AT_CMD(data,2)=XBEE_AT_MY;
        XBEE_AT_CMD(data,4)=0xFFFF;
        Link_SendATCommand(0,data,6,!LINK_IGNORE_RADIO_1);

        //Channel scan mask. Channel 0x0B is reserved
        //for core network use only, scanning 4 channels to make connection faster
        XBEE_AT_CMD(data,2)=XBEE_AT_SC;
        XBEE_AT_CMD(data,4)=0x0010;//0xFEFF;//NOTE: Little endian architecture. 1,6,11,15 for best spread!!!!
        Link_SendATCommand(0,data,6,!LINK_IGNORE_RADIO_1);

        //Coordinator association options.
        XBEE_AT_CMD(data,2)=XBEE_AT_A2;
        data[4]=XBEE_A2_REASSIGN_ID |
                XBEE_A2_REASSIGN_CH |
                XBEE_A2_ALLOW_ASSOCIATION;
        Link_SendATCommand(0,data,5,!LINK_IGNORE_RADIO_1);

        //Enable coordinator.
        XBEE_AT_CMD(data,2)=XBEE_AT_CE;
        data[4]=1;
        Link_SendATCommand(0,data,5,!LINK_IGNORE_RADIO_1);
        
        DEBUG_OFF(1);

#else
    //Bring up host interface.
    DEBUG_ON(0);
        
    //Initialize to specified baud rate, 8N1.
    UBRR0=BAUD(LINK_BAUD_RATE);
        
    //Enable RX and TX with RX interrupt only.
    //The TX (DRE) interrupt is enabled when sending data.
    UCSR0B=_BV(RXEN0) | _BV(TXEN0) | _BV(RXCIE0);

    //Reset module to clear previous settings.
    Link_ResetModule(0);

    //Module local address. Must be set to 0xFFFF
    //to enable 64-bit addressing.
    XBEE_AT_CMD(data,2)=XBEE_AT_MY;
    XBEE_AT_CMD(data,4)=0xFFFF;
    Link_SendATCommand(0,data,6,!LINK_IGNORE_RADIO_0);

    //Channel scan mask. Channel 0x0B is reserved
    //for the core.
    XBEE_AT_CMD(data,2)=XBEE_AT_SC;
    XBEE_AT_CMD(data,4)=0x0010;//0xFEFF;//NOTE: Little endian architecture.
    Link_SendATCommand(0,data,6,!LINK_IGNORE_RADIO_0);

    //End device association options.
    XBEE_AT_CMD(data,2)=XBEE_AT_A1;
    data[4]=XBEE_A1_AUTO_ASSOCIATE |
            XBEE_A1_REASSIGN_ID |
            XBEE_A1_REASSIGN_CH;
    Link_SendATCommand(0,data,5,!LINK_IGNORE_RADIO_0);

    DEBUG_OFF(0);
#endif
    
    DEBUG_OFF_BLOCK(0x70);
}

uint8_t Link_SendFrame(uint8_t interface,
                       MACAddress* destination,
                       uint8_t* data,
                       uint8_t length)
{
    uint8_t i, dataIndex;

    //Make sure the payload fits into the frame.
    //If not, discard the payload. In addition,
    //check that the payload meets XBee specifications.
    if(length>LINK_MTU_SIZE || length>XBEE_MAX_PAYLOAD_SIZE)return 0;

    //API type identifier.
    link_txPayloadBuffer[LINK_TX_POS_API_ID]=XBEE_API_TX_64;

    //Frame ID 0 (disable response frame).
    link_txPayloadBuffer[LINK_TX_POS_FRAME]=0;
    
    //Destination address.
#if FORCE_BROADCAST
    destination->addr64=LINK_ADDR_BROADCAST;
#endif
    link_txPayloadBuffer[LINK_TX_POS_DEST+0]=destination->addr8.byte7;
    link_txPayloadBuffer[LINK_TX_POS_DEST+1]=destination->addr8.byte6;
    link_txPayloadBuffer[LINK_TX_POS_DEST+2]=destination->addr8.byte5;
    link_txPayloadBuffer[LINK_TX_POS_DEST+3]=destination->addr8.byte4;
    link_txPayloadBuffer[LINK_TX_POS_DEST+4]=destination->addr8.byte3;
    link_txPayloadBuffer[LINK_TX_POS_DEST+5]=destination->addr8.byte2;
    link_txPayloadBuffer[LINK_TX_POS_DEST+6]=destination->addr8.byte1;
    link_txPayloadBuffer[LINK_TX_POS_DEST+7]=destination->addr8.byte0;

    //Options (none).
    link_txPayloadBuffer[LINK_TX_POS_OPTIONS]=0x00;

    //Packet data.
    //Offset length and i to save operation time.
    length+=LINK_TX_POS_DATA;
    dataIndex=0;
    for(i=LINK_TX_POS_DATA;i<length;i++)
    {
        link_txPayloadBuffer[i]=data[dataIndex++];
    }

    return Link_SendAPIFrame(interface,link_txPayloadBuffer,length);
}

void Link_SendAIRequest(uint8_t interface)
{
    uint8_t data[4];
    data[0]=XBEE_API_AT_COMMAND;//API identifier.
    data[1]=1;//Frame ID (0 to disable response).
    XBEE_AT_CMD(data,2)=XBEE_AT_AI;//Assocation indication request.
    Link_SendAPIFrame(interface,data,4);
}

uint8_t Link_ReceiveFrame(uint8_t interface, LinkFrame* frame)
{
    unsigned char i;
//    char led = 0xff;
    //It is sufficient to check only that the
    //RX queue has data since corrupt frames
    //are discarded.
    if(Link_FramePending(interface))
    {
        
//		PORTC = ~led|0x01;
		
		//Retrieve link frame type.
        frame->type=link_rxQueue[interface][LINK_INDEX_OFFSET(link_rxHead[interface],
                                                              XBEE_POS_API_ID,
                                                              LINK_RX_INDEX_MASK)];
        
        //Retrieve payload length.
        frame->length=Link_PayloadLength(interface);
        
        //Copy head frame to receive buffer.
        for(i=XBEE_POS_FRAME_PAYLOAD;i<XBEE_POS_FRAME_PAYLOAD+frame->length;i++)
        {
            link_rxPayloadBuffer[i-XBEE_POS_FRAME_PAYLOAD]=link_rxQueue[interface][LINK_INDEX_OFFSET(link_rxHead[interface],
                                                                                               i,
                                                                                               LINK_RX_INDEX_MASK)];
        }

        switch(frame->type)
        {
        case XBEE_API_MODEM_STATUS:
            //Set payload pointer.
            frame->payload=link_rxPayloadBuffer+LINK_STATUS_POS_DATA;
            link_moduleStatus[interface]=link_rxPayloadBuffer[LINK_STATUS_POS_DATA];
            //!\todo Reset interface (recall Link_Init) if module reset.
            break;
        case XBEE_API_AT_RESPONSE:
            //Set command type value.
            frame->commandType=*(uint16_t*)(link_rxPayloadBuffer+LINK_AT_POS_CMD);
            
            pendingCmd.valid=1;
            pendingCmd.type=frame->commandType;
            pendingCmd.id=link_rxPayloadBuffer[LINK_AT_POS_ID];
            pendingCmd.response=link_rxPayloadBuffer[LINK_AT_POS_STATUS];

            //Is this an AI command response? If not,
            //discard the frame.
            if(frame->commandType==XBEE_AT_AI)
            {
                //Set payload pointer.
                frame->payload=link_rxPayloadBuffer+LINK_AT_POS_DATA;
            }
            else frame->type=LINK_TYPE_NONE;

            break;
        case XBEE_API_RX_64:
            //Set payload pointer.
            frame->payload=link_rxPayloadBuffer+LINK_RX_POS_DATA;

            //Set source address.
            frame->source.addr8.byte7=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 0];
            frame->source.addr8.byte6=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 1];
            frame->source.addr8.byte5=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 2];
            frame->source.addr8.byte4=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 3];
            frame->source.addr8.byte3=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 4];
            frame->source.addr8.byte2=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 5];
            frame->source.addr8.byte1=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 6];
            frame->source.addr8.byte0=link_rxPayloadBuffer[LINK_RX_POS_SOURCE + 7];

            //Set RSSI.
            frame->rssi=link_rxPayloadBuffer[LINK_RX_POS_RSSI];

            //Set options.
            frame->options=link_rxPayloadBuffer[LINK_RX_POS_OPTIONS];
            break;
#if LINK_IGNORE_RADIO_0 || (LINK_NUM_INTERFACES==2 && LINK_IGNORE_RADIO_1)
        case XBEE_API_TX_64:
            //NOTE: This case exists for debug purposes,
            //and is only active when one of the radio
            //module ignore flags is enabled.
            
            //Set payload pointer.
            frame->payload=link_rxPayloadBuffer+LINK_TX_POS_DATA-1;

            //Set source address.
            frame->source.addr8.byte7=link_rxPayloadBuffer[LINK_TX_POS_DEST + 0-1];
            frame->source.addr8.byte6=link_rxPayloadBuffer[LINK_TX_POS_DEST + 1-1];
            frame->source.addr8.byte5=link_rxPayloadBuffer[LINK_TX_POS_DEST + 2-1];
            frame->source.addr8.byte4=link_rxPayloadBuffer[LINK_TX_POS_DEST + 3-1];
            frame->source.addr8.byte3=link_rxPayloadBuffer[LINK_TX_POS_DEST + 4-1];
            frame->source.addr8.byte2=link_rxPayloadBuffer[LINK_TX_POS_DEST + 5-1];
            frame->source.addr8.byte1=link_rxPayloadBuffer[LINK_TX_POS_DEST + 6-1];
            frame->source.addr8.byte0=link_rxPayloadBuffer[LINK_TX_POS_DEST + 7-1];

            //Set RSSI.
            frame->rssi=link_rxPayloadBuffer[LINK_TX_POS_FRAME-1];

            //Set options.
            frame->options=link_rxPayloadBuffer[LINK_TX_POS_OPTIONS-1];

            frame->type=XBEE_API_RX_64;
            break;
#endif
        default:
            frame->type=LINK_TYPE_NONE;
            break;
        }

        //Frame is no longer needed in receive queue.
        //Discard data and update head pointer.
        link_rxBytesPending[interface]-=Link_FrameLength(interface);
        link_rxHead[interface]=LINK_INDEX_OFFSET(link_rxHead[interface],
                                                 Link_FrameLength(interface),
                                                 LINK_RX_INDEX_MASK);

        return frame->type;
    }
    else
    {
		//PORTC = ~led|0x04;
		return LINK_TYPE_NONE;
    }
}

void Link_EndianSwap(MACAddress* address)
{
    uint8_t b0, b1, b2, b3;
    
    b0=address->addr8.byte0;
    b1=address->addr8.byte1;
    b2=address->addr8.byte2;
    b3=address->addr8.byte3;

    address->addr8.byte0=address->addr8.byte7;
    address->addr8.byte1=address->addr8.byte6;
    address->addr8.byte2=address->addr8.byte5;
    address->addr8.byte3=address->addr8.byte4;
    address->addr8.byte4=b3;
    address->addr8.byte5=b2;
    address->addr8.byte6=b1;
    address->addr8.byte7=b0;
}

/*inline*/ void Link_Execute(void)
{
    //Has this interface been inactive? If so,
    //revert the pending frame.
    if(link_activityTimer[0]>=LINK_INACTIVITY_TIMEOUT && link_rxLength[0]>0)
    {
        link_rxInsertionPoint[0]=link_rxTail[0];
        link_rxLength[0]=0;
    }

/*#if LINK_NUM_INTERFACES==2
    //Has this interface been inactive? If so,
    //revert the pending frame.
    if(link_activityTimer[1]>=LINK_INACTIVITY_TIMEOUT && link_rxLength[1]>0)
    {
        link_rxInsertionPoint[1]=link_rxTail[1];
        link_rxLength[1]=0;
    }
#endif*/
}

/*inline*/ void Link_UpdateTimers(void)
{
/*#if ENABLE_DEBUG_LEDS
    static uint16_t tick=0;
    if(++tick==1000)
    {
        tick=0;
        DEBUG_TOGGLE(7);
    }
#endif*/
    
    if(link_activityTimer[0]<255)link_activityTimer[0]++;
/*#if LINK_NUM_INTERFACES==2
    if(link_activityTimer[1]<255)link_activityTimer[1]++;
#endif*/
}

void Link_ResetModule(unsigned char interface)
{
    LinkFrame frame;
    uint8_t data[4];
    
    link_moduleStatus[0]=-1;

    //Issue reset command.
    data[0]=XBEE_API_AT_COMMAND;//API identifier.
    data[1]=0;//Frame ID (0 to disable response).
    XBEE_AT_CMD(data,2)=XBEE_AT_FR;
/*#if LINK_NUM_INTERFACES==2
    if(interface==0)Link_SendATCommand(0,data,4,!LINK_IGNORE_RADIO_0);
    else Link_SendATCommand(1,data,4,!LINK_IGNORE_RADIO_1);
#else*/
    Link_SendATCommand(0,data,4,!LINK_IGNORE_RADIO_0);
//#endif

    //Wait for reset completion.
/*#if LINK_NUM_INTERFACES==2
    if((interface==0 && !LINK_IGNORE_RADIO_0) ||
       (interface==1 && !LINK_IGNORE_RADIO_1))
    {
#else*/
    if(!LINK_IGNORE_RADIO_0)
    {
//#endif
        while(link_moduleStatus[interface]<0)
        {
            Link_ReceiveFrame(interface,&frame);
            Link_Execute();
        }
    }
}

void Link_SendATCommand(uint8_t interface, uint8_t* data, uint8_t length, uint8_t wait)
{
    LinkFrame frame;
    static uint8_t nextID=1;
    
    //SEt frame ID (0 to disable response).
    if(wait)
    {
        data[1]=nextID;
        if(++nextID==0)nextID=1;
    }
    else data[1]=0;

    //Send first attempt.
    pendingCmd.valid=0;
    link_activityTimer[interface]=0;
    Link_SendAPIFrame(interface,data,length);

    //Are we supposed to wait for the response?
    //If not, just wait until the command sends.
    if(!wait)
    {
        if(interface==0)while(!LINK_USART_READY(0));
        else while(!LINK_USART_READY(1));
        return;
    }

    //Wait for successful command response.
    while(!pendingCmd.valid ||
          pendingCmd.id!=data[1] ||
          pendingCmd.response!=0)
    {
        //Did we get the response we wanted or
        //did the current attempt expire?
        if((pendingCmd.valid && pendingCmd.id==data[1]) ||
           link_activityTimer[interface]>=LINK_AT_RESPONSE_TIMEOUT)
        {
            data[1]=nextID;
            if(++nextID==0)nextID=1;
            pendingCmd.valid=0;
            link_activityTimer[interface]=0;
            Link_SendAPIFrame(interface,data,length);
        }
        //Did we get an unwanted response?
        else if(pendingCmd.valid)pendingCmd.valid=0;
        
        Link_ReceiveFrame(interface,&frame);
        Link_Execute();
    }
}

uint8_t Link_SendAPIFrame(uint8_t interface, uint8_t* data, uint8_t length)
{
    uint8_t i, checksum=0xFF;
    volatile uint16_t *head=&link_txHead[interface],
                      *tail=&link_txTail[interface];

    //Can this frame fit in the buffer? We should never
    //send partial frames to the radio to avoid having
    //to wait for timeouts.
    if(length+4>link_txRemaining[interface])return 0;
    link_txRemaining[interface]-=(length+4);//Allocate space in buffer.

    //Start byte. All API messages must start with 0x7E.
    link_txQueue[interface][*head]=0x7E;
    LINK_INDEX_INC(*head,LINK_TX_INDEX_MASK);

    //Message length. No message should be larger than 100B.
    link_txQueue[interface][*head]=0x00;//MSB.
    LINK_INDEX_INC(*head,LINK_TX_INDEX_MASK);
    link_txQueue[interface][*head]=length;//LSB (length includes API identifier).
    LINK_INDEX_INC(*head,LINK_TX_INDEX_MASK);
    
    //Copy each byte into TX buffer.
    for(i=0;i<length;i++)
    {
        link_txQueue[interface][*head]=data[i];
        LINK_INDEX_INC(*head,LINK_TX_INDEX_MASK);
        checksum-=data[i];
    }

    //Checksum is 0xFF-SUM(data).
    link_txQueue[interface][*head]=checksum;
    LINK_INDEX_INC(*head,LINK_TX_INDEX_MASK);

    //Check if USART is ready to receive data.
    //If so, send first byte to initiate transmission.
    //NOTE: The order of operations is very important here.
    //      The tail pointer must be incremented BEFORE
    //      enabling the DRE interrupt, or the interrupt will
    //      fire early, causing a byte to be skipped.
#if AVR_NUM_INTERFACES==2
    if(interface==0)
    {
        if(LINK_USART_READY(0))
        {
            link_txRemaining[0]++;
            UDR0=link_txQueue[0][*tail];
            LINK_INDEX_INC(*tail,LINK_TX_INDEX_MASK);
            UCSR0B|=_BV(UDRIE0);
        }
    }
/*#if LINK_NUM_INTERFACES==2
    else
    {
        if(LINK_USART_READY(1))
        {
            link_txRemaining[1]++;
            UDR1=link_txQueue[1][*tail];
            LINK_INDEX_INC(*tail,LINK_TX_INDEX_MASK);
            UCSR1B|=_BV(UDRIE1);
        }
    }
#endif*/
#else
    if(LINK_USART_READY(0))
    {
        link_txRemaining[0]++;
        UDR=link_txQueue[0][*tail];
        LINK_INDEX_INC(*tail,LINK_TX_INDEX_MASK);
        UCSRB|=_BV(UDRIE);
    }
#endif

    return 1;
}

inline uint8_t Link_FramePending(uint8_t interface)
{
    return link_rxBytesPending[interface]>0;
}

inline uint8_t Link_FrameLength(uint8_t interface)
{
    //All RX frames have 5 bytes of overhead.
    //Start byte - 1B
    //Length     - 2B
    //API ID     - 1B
    //Checksum   - 1B
    return Link_PayloadLength(interface)+5;
}

inline uint8_t Link_PayloadLength(uint8_t interface)
{
    //No payload is larger than 255 bytes, so
    //it is sufficient to read the length LSB only.
    //The returned payload length does not include the API ID.
    return link_rxQueue[interface][LINK_INDEX_OFFSET(link_rxHead[interface],
                                                     XBEE_POS_LENGTH_LO,
                                                     LINK_RX_INDEX_MASK)]-1;
}

#if AVR_NUM_INTERFACES==2
ISR(USART0_RX_vect)
{
    uint8_t input=UDR0;
    uint16_t space=LINK_RX_QUEUE_SIZE-link_rxBytesPending[0];
    uint8_t payloadLength;

    //Check for start byte.
    if(link_rxLength[0]==0)
    {
        if(input!=0x7E)return;
        else link_rxChecksum[0]=0xFF;
    }
    
    link_activityTimer[0]=0;

    payloadLength=link_rxQueue[0][LINK_INDEX_OFFSET(link_rxTail[0],
                                                    XBEE_POS_LENGTH_LO,
                                                    LINK_RX_INDEX_MASK)]+4;

    //Is there room in the RX queue for the
    //incoming data? If not, revert frame.
    if((link_rxLength[0]<3 && space<4) ||
       (link_rxLength[0]>=3 && space<payloadLength))
    {
        link_rxInsertionPoint[0]=link_rxTail[0];
        link_rxLength[0]=0;
        return;
    }

    //Place byte in RX queue at insertion point
    //and update receive length and checksum.
    link_rxQueue[0][link_rxInsertionPoint[0]]=input;
    LINK_INDEX_INC(link_rxInsertionPoint[0],LINK_RX_INDEX_MASK);
    link_rxLength[0]++;
    if(link_rxLength[0]>3)link_rxChecksum[0]-=input;

    //Is this message done?
    if(link_rxLength[0]>3 &&
       link_rxLength[0]==payloadLength)
    {
        //Is this a valid message? If so, update tail
        //pointer and received byte count. If not,
        //revert insertion point.
        if(link_rxChecksum[0]==0x00)
        {
            link_rxTail[0]=link_rxInsertionPoint[0];
            link_rxBytesPending[0]+=link_rxLength[0];
        }
        else
        {
            link_rxInsertionPoint[0]=link_rxTail[0];
        }
        link_rxLength[0]=0;
    }
}

ISR(USART0_UDRE_vect)
{
    //Is there more data to be sent? If not,
    //disable the interrupt.
    if(link_txRemaining[0]!=LINK_TX_QUEUE_SIZE)
    {
        link_txRemaining[0]++;
        UDR0=link_txQueue[0][link_txTail[0]];
        LINK_INDEX_INC(link_txTail[0],LINK_TX_INDEX_MASK);
    }
    else UCSR0B&=~_BV(UDRIE0);
}

/*#if LINK_NUM_INTERFACES==2
#if DEVICE_TYPE==1
ISR(USART1_RX_vect)
{
    uint8_t input=UDR1;
    uint16_t space=LINK_RX_QUEUE_SIZE-link_rxBytesPending[1];
    uint8_t payloadLength;

    //Check for start byte.
    if(link_rxLength[1]==0)
    {
        if(input!=0x7E)return;
        else link_rxChecksum[1]=0xFF;
    }
    
    link_activityTimer[1]=0;

    payloadLength=link_rxQueue[1][LINK_INDEX_OFFSET(link_rxTail[1],
                                                    XBEE_POS_LENGTH_LO,
                                                    LINK_RX_INDEX_MASK)]+4;

    //Is there room in the RX queue for the
    //incoming data? If not, revert frame.
    if((link_rxLength[1]<3 && space<4) ||
       (link_rxLength[1]>=3 && space<payloadLength))
    {
        link_rxInsertionPoint[1]=link_rxTail[1];
        link_rxLength[1]=0;
        return;
    }

    //Place byte in RX queue at insertion point
    //and update receive length and checksum.
    link_rxQueue[1][link_rxInsertionPoint[1]]=input;
    LINK_INDEX_INC(link_rxInsertionPoint[1],LINK_RX_INDEX_MASK);
    link_rxLength[1]++;
    if(link_rxLength[1]>3)link_rxChecksum[1]-=input;

    //Is this message done?
    if(link_rxLength[1]>3 &&
       link_rxLength[1]==payloadLength)
    {
        //Is this a valid message? If so, update tail
        //pointer and received byte count. If not,
        //revert insertion point.
        if(link_rxChecksum[1]==0x00)
        {
            link_rxTail[1]=link_rxInsertionPoint[1];
            link_rxBytesPending[1]+=link_rxLength[1];
        }
        else
        {
            link_rxInsertionPoint[1]=link_rxTail[1];
        }
        link_rxLength[1]=0;
    }
}

ISR(USART1_UDRE_vect)
{
    //Is there more data to be sent? If not,
    //disable the interrupt.
    if(link_txRemaining[1]!=LINK_TX_QUEUE_SIZE)
    {
        link_txRemaining[1]++;
        UDR1=link_txQueue[1][link_txTail[1]];
        LINK_INDEX_INC(link_txTail[1],LINK_TX_INDEX_MASK);
    }
    else UCSR1B&=~_BV(UDRIE1);
}
#endif
#endif*/
#else
//!\todo Implement single-interface device interrupts.
#endif
