/************************************************************
* This file is part of the XDN Router project.
*
* XDN Router is a packet routing device to be used with
* a local XDN (XBee Distributed Network) configuration.
* It supports both core and edge routing modes.
* Copyright (C) 2008 Adam Shapiro, Cornell University
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
************************************************************/
/** \file
 * \brief Public link layer declarations.
 *
 * This file contains public declarations for the
 * XDN link layer.
 */

#ifndef __link_h__
#define __link_h__

#include <stdint.h>
#include "router.h"

/**
 * \brief A 64-bit MAC address for a given device.
 *
 * Each device interface has a factory-assigned
 * 64-bit Media Access Control address that
 * uniquely identifies the device on the network.
 */
union basic_MACAddress
{
    struct
    {
        uint8_t byte0;
        uint8_t byte1;
        uint8_t byte2;
        uint8_t byte3;
        uint8_t byte4;
        uint8_t byte5;
        uint8_t byte6;
        uint8_t byte7;
    } addr8;
    struct
    {
        uint32_t lo;
        uint32_t hi;
    } addr32;
    uint64_t addr64;
};

/**
 * \brief An alias for basic_MACAddress.
 */
typedef union basic_MACAddress MACAddress;

/**
 * \brief The link layer frame structure.
 *
 * This structure is filled and returned
 * by the Link_ReceiveFrame function when
 * a frame is available.
 */
struct basic_LinkFrame
{
    uint8_t type;      //!< Frame type. See \ref LinkTypes for more details.
    uint8_t length;    //!< Frame payload length.
    uint8_t* payload;  //!< Frame payload.
    MACAddress source; //!< Frame source address (\ref LINK_TYPE_RX only).
    union
    {
        struct
        {
            uint8_t rssi;     //!< Receive Signal Strength Indicator in units of -dBm. (\ref LINK_TYPE_RX only).
            uint8_t options;  //!< Frame options (\ref LINK_TYPE_RX only). See \ref LinkRXOptions for more details.
        };
        uint16_t commandType; //!< Command type used for AT command responses.
    };
};

/**
 * \brief An alias for basic_LinkFrame.
 */
typedef struct basic_LinkFrame LinkFrame;

/**
 * \brief Maximum frame size.
 *
 * The maximum size of a single frame in bytes.
 * This value includes both payload and frame
 * overhead. Because of limitations of the XBee
 * modules, this value should never be larger
 * than 256 bytes.
 *
 * Additionally, frame payloads should never
 * exceed \ref XBEE_MAX_PAYLOAD_SIZE.
 *
 * Note: This value should always be a power of two.
 */
#define LINK_MTU_SIZE 64

/**
 * \brief Broacast link address.
 *
 * Frames sent to this address are broadcast
 * to all devices on a local network.
 */
#define LINK_ADDR_BROADCAST 0x000000000000FFFF

/**
 * \defgroup LinkRXOptions Link RX Frame Options
 * \brief Options for a link RX frame.
 *
 * Transmit frames may be broadcast to any address
 * or, additionally, to any PAN ID.
 */
//@{
#define LINK_RX_OPTION_ADDR_BROADCAST 0x02 //!< Address broadcast enabled.
#define LINK_RX_OPTION_PAN_BROADCAST  0x04 //!< PAN broadcast enabled.
//@}

/**
 * \defgroup LinkTypes Link Types
 * \brief Link frame types.
 *
 * Type identifier values (1B) for link frames
 * recevied from an XBee module. See \ref APITypes
 * for more details.
 */
//@{
#define LINK_TYPE_NONE      0xFF //!< No frame available.
#define LINK_TYPE_STATUS    0x8A //!< Modem status. See \ref LinkStatus for more details.
#define LINK_TYPE_ASSOC_IND 0x88 //!< Association indication response. See \ref LinkIndicators for more details.
#define LINK_TYPE_RX        0x80 //!< Receive frame (64-bit addressing only).
//@}

/**
 * \defgroup LinkStatus Link Status Values
 * \brief Possible values for link status frames.
 *
 * Link status frames provide indications of the link
 * state in real time from the XBee module. Status frames
 * are used to indicate association/disconnection, loss
 * of synchronization, and module reset.
 */
//@{
#define LINK_STATUS_HW_RESET                0
#define LINK_STATUS_WATCHDOG_RESET          1
#define LINK_STATUS_ASSOCIATED              2
#define LINK_STATUS_DISASSOCIATED           3
#define LINK_STATUS_SYNC_LOST               4
#define LINK_STATUS_COORDINATOR_REALIGNMENT 5
#define LINK_STATUS_COORDINATOR_STARTED     6
//@}

/**
 * \defgroup LinkIndicators Link association indicator values.
 * \brief Possible values for link association indication frames.
 *
 * Link association indication frames provide indications
 * of the link state in real time from the XBee module.
 * Assocation indicators provide further detail than \ref LinkStatus
 * for link association status.
 */
//@{
#define LINK_AI_SUCCESSFUL_COMPLETION 0x00
#define LINK_AI_ACTIVE_SCAN_TIMEOUT   0x01
#define LINK_AI_NO_PANS_FOUND         0x02
#define LINK_AI_ASSOC_FORBIDDEN       0x03
#define LINK_AI_NO_BEACON_SUPPORT     0x04
#define LINK_AI_PAN_ID_MISMATCH       0x05
#define LINK_AI_CHANNEL_MISMATCH      0x06
#define LINK_AI_ENERGY_SCAN_TIMEOUT   0x07
#define LINK_AI_COORD_START_FAILED    0x08
#define LINK_AI_COORD_BAD_PARAM       0x09
#define LINK_AI_COORD_REALIGNING      0x0A
#define LINK_AI_ASSOC_REQ_NOT_SENT    0x0B
#define LINK_AI_ASSOC_REQ_TIMEOUT     0x0C
#define LINK_AI_ASSOC_REQ_BAD_PARAM   0x0D
#define LINK_AI_ASSOC_REQ_CCA_FAIL    0x0E
#define LINK_AI_NO_ACK_FROM_COORD     0x0F
#define LINK_AI_NO_REPLY_FROM_COORD   0x10
#define LINK_AI_SYNC_LOSS             0x12
#define LINK_AI_DISASSOCIATED         0x13
#define LINK_AI_ASSOCIATING           0xFF
//@}

/**
 * \brief Initialize specified interface..
 *
 * This function assumes a mostly-default (factory)
 * configuration for the XBee module. In particular, it
 * assumes the module is set to 9600 Baud 8N1.
 * The only dif`ference from factory conditions expected
 * is that API mode is enabled.
 *
 * This function requires interrupts to be enabled.
 * Be sure to call sei prior to calling Link_Init.
 *
 * \param interface is the unique identifier for the interface.
 */
void Link_Init(uint8_t interface);

/**
 * \brief Send packet.
 *
 * Format packet into an XBee TX frame and send it on the interface.
 *
 * \param interface is the unique identifier for the interface.
 * \param destination is the 64-bit hardware MAC address for the destination device.
 * \param data is a pointer to the packet to be sent.
 * \param length is the length of the packet to be sent.
 *
 * \return Returns 1 if succeeded, 0 otherwise.
 */
uint8_t Link_SendFrame(uint8_t interface,
                       MACAddress* destination,
                       uint8_t* data,
                       uint8_t length);

/**
 * \brief Send association indication request.
 *
 * Send an assocation indication request command
 * to the XBee module.
 *
 * \param interface is the unique identifier for the interface.
 */
void Link_SendAIRequest(uint8_t interface);

/**
 * \brief Get the next available received frame.
 *
 * Returns a pointer to the payload of the frame
 * currently at the head of the interface receive
 * queue.
 *
 * The payload of the next available frame is copied
 * into a receive buffer, and the data is discarded
 * from the receive queue.
 *
 * For receive frames, the source address and
 * RSSI reading from the frame are provided for
 * further processing.
 *
 * \param interface is the unique identifier for the interface.
 * \param frame is a pointer to a \ref LinkFrame structure to be
 *        filled based on the frame type.
 *
 * \return Returns the type of the frame. Returns \ref LINK_TYPE_NONE
 *         if no frame is available. See \ref LinkTypes for more details.
 */
uint8_t Link_ReceiveFrame(uint8_t interface, LinkFrame* frame);

/**
 * \brief Change MAC address endianness.
 *
 * AVRs are a little endian architecture, however
 * the current standard for networking protocols is
 * to use big endian. This function converts a little
 * endian MAC address to big endian, and vice versa.
 *
 * \param address is a pointer to the address to be converted.
 */
void Link_EndianSwap(MACAddress* address);

/**
 * \brief Link layer execute function.
 *
 * This function should be called frequently
 * to check link layer activity timers.
 */
/*inline */void Link_Execute(void);

/**
 * \brief Update internal timers.
 *
 * This function must be called every 1 ms
 * to update any timing-critical timers used
 * by the link layer.
 *
 * This function is timing-sensitive and must
 * be called in an interrupt for proper operation.
 */
/*inline */void Link_UpdateTimers(void);

#endif
