/************************************************************
* This file is part of the XDN Router project.
*
* XDN Router is a packet routing device to be used with
* a local XDN (XBee Distributed Network) configuration.
* It supports both core and edge routing modes.
* Copyright (C) 2008 Adam Shapiro, Cornell University
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
************************************************************/
/** \file
 * \brief Private link layer declarations.
 *
 * This file contains private declarations for the
 * XDN link layer. Its contents are used only by the
 * underlying XDN code, and are not intended for
 * user use. Modify with extreme caution.
 */

#ifndef __link_defs_h__
#define __link_defs_h__

#include "link_interfaces.h"

/**
 * \brief A pending AT command response.
 *
 * Each AT command response has an 8-bit
 * command ID, a 16-bit command type, and
 * an 8-bit response value (success or failure).
 */
struct basic_PendingATResponse
{
    uint8_t valid; //!< Response valid.
    uint8_t id; //!< Command ID.
    uint16_t type; //!< Command type.
    uint8_t response; //!< Response (0=Success, 1=Failure).
};

/**
 * \brief An alias for basic_PendingATResponse.
 */
typedef struct basic_PendingATResponse PendingATResponse;

/**
 * \brief Maximum XBee frame size.
 *
 * The maximum size of a frame payload in bytes.
 * This value is determined by the maximum packet
 * size in the XBee specification document (pg. 50)
 * and should not be modified.
 *
 * This number is offset by 4 bytes to account for
 * the link layer header.
 */
#define XBEE_MAX_PAYLOAD_SIZE 96

/**
 * \brief Maximum number of frames in the TX queue.
 *
 * This parameter specifies the number of maximum-sized
 * frames that can exist in the transmit queue at any
 * time. This parameter multiplied by \ref LINK_MTU_SIZE
 * yields the total transmit queue size. See \ref LINK_TX_QUEUE_SIZE
 * for more details.
 *
 * Note: This value should always be a power of two.
 */
#define LINK_TX_MAX_FRAMES 4

/**
 * \brief TX queue size.
 *
 * This parameter specifies the transmit queue size
 * in bytes for each given interface.
 *
 * The buffer size is calculated as:
 *
 *    Size (B) = \ref LINK_TX_MAX_FRAMES * \ref LINK_MTU_SIZE
 *
 * The actual number of frames in the queue at a
 * given time can vary depending on the size of the
 * pending frames.
 *
 * The transmit queue is a First-In, First-Out (FIFO) queue.
 * Frames are transmitted in the order in which they are enqueued.
 *
 * The queue is implemented as a circular buffer and so,
 * to fascilitate indexing, its size must always be a power of 2.
 */
#define LINK_TX_QUEUE_SIZE (LINK_TX_MAX_FRAMES * LINK_MTU_SIZE)

/**
 * \brief TX queue index mask.
 *
 * This value is used to mask the transmit queue
 * index value in order to wrap around after reaching
 * the end of the queue.
 */
#define LINK_TX_INDEX_MASK (LINK_TX_QUEUE_SIZE - 1)

/**
 * \brief Maximum number of frames in the RX queue.
 *
 * This parameter specifies the number of maximum-sized
 * frames that can exist in the receive queue at any
 * time. This parameter multiplied by \ref LINK_MTU_SIZE
 * yields the total receive queue size. See \ref LINK_RX_QUEUE_SIZE
 * for more details.
 *
 * Note: This value should always be a power of two.
 */
#define LINK_RX_MAX_FRAMES 2

/**
 * \brief RX queue size.
 *
 * This parameter specifies the receive queue size
 * in bytes for each given interface.
 *
 * The buffer size is calculated as:
 *
 *    Size (B) = \ref LINK_RX_MAX_FRAMES * \ref LINK_MTU_SIZE
 *
 * The actual number of frames in the queue at a
 * given time can vary depending on the size of the
 * pending frames.
 *
 * The receive queue is a First-In, First-Out (FIFO) queue.
 * Frames are processed in the order in which they are received.
 *
 * The queue is implemented as a circular buffer and so,
 * to fascilitate indexing, its size must always be a power of 2.
 */
#define LINK_RX_QUEUE_SIZE (LINK_RX_MAX_FRAMES * LINK_MTU_SIZE)

/**
 * \brief RX queue index mask.
 *
 * This value is used to mask the receive queue
 * index value in order to wrap around after reaching
 * the end of the queue.
 */
#define LINK_RX_INDEX_MASK (LINK_RX_QUEUE_SIZE - 1)

/**
 * \brief Link layer baud rate.
 *
 * This parameter specifies the baud rate used to
 * communicate with the XBee module. The value used
 * must be one of the available options as specified
 * in \ref BaudRates.
 */
#define LINK_BAUD_RATE 38400 //9600//

/**
 * \brief Link layer activity timeout.
 *
 * This parameter defines the timeout for data in
 * the receiver buffer. Pending data is discarded
 * after this time (in ms) expires.
 */
#define LINK_INACTIVITY_TIMEOUT 2

/**
 * \brief AT command response timeout.
 *
 * This parameter defines the timeout (in ms)
 * after which AT commands are considered lost.
 */
#define LINK_AT_RESPONSE_TIMEOUT 50

/**
 * \brief Baud rate selection macro.
 *
 * This macro selects the appropriate AT baud rate
 * command parameter for the communication rate
 * specified in \ref LINK_BAUD_RATE. See \ref BaudRates
 * for more details.
 */
#define LINK_BAUD_RATE_PARAM(x) LINK_BAUD_RATE_PARAM_CAT(x)
#define LINK_BAUD_RATE_PARAM_CAT(x) (XBEE_BAUD_ ## x)

/**
 * \defgroup XBeePos XBee Frame Structure
 * \brief Component positions for all XBee frames.
 *
 * All XBee frames have a start byte, a 16-bit
 * payload length, and a 1B checksum.
 */
//@{
#define XBEE_POS_START_BYTE 0    //!< Start byte (0x7E).
#define XBEE_POS_LENGTH_HI  1    //!< Length MSB.
#define XBEE_POS_LENGTH_LO  2    //!< Length LSB.
#define XBEE_POS_FRAME_DATA 3    //!< Frame data (including API type).
#define XBEE_POS_FRAME_PAYLOAD 4 //!< Frame payload.
#define XBEE_POS_API_ID XBEE_POS_FRAME_DATA //!< API ID. See \ref APITypes for more details.
//@}

/**
 * \defgroup APITypes API Types
 * \brief XBee API message types.
 *
 * API identifier values (1B) as defined by the XBee
 * specification (pg. 54). The size and format of the
 * payload of a link frame depends on the type of frame
 * received.
 */
//@{
#define XBEE_API_MODEM_STATUS 0x8A //!< Modem status.
#define XBEE_API_AT_COMMAND   0x08 //!< Issue AT command. See \ref ATCommands for more details.
#define XBEE_API_AT_QUEUE     0x09 //!< Queue AT command (must be applied later). See \ref ATCommands for more details.
#define XBEE_API_AT_RESPONSE  0x88 //!< AT command response.
#define XBEE_API_TX_64        0x00 //!< Transmit request (64-bit addressing).
#define XBEE_API_TX_16        0x01 //!< Transmit request (16-bit addressing).
#define XBEE_API_TX_STATUS    0x89 //!< Transmit status.
#define XBEE_API_RX_64        0x80 //!< Receive frame (64-bit addressing).
#define XBEE_API_RX_16        0x81 //!< Receive frame (16-bit addressing).
//@}

/**
 * \defgroup ATCommands AT Commands
 * \brief XBee AT commands.
 *
 * AT command values (2B) as defined by the XBee specification.
 * Most commands correspond to control registers within the XBee
 * module that may be either read by issuing them with no payload,
 * or set by issuing them with associating values.
 *
 * Note that values sent to the module are specified as base-16 integers.
 * A number of commands have associated parameter values. For more details,
 * refer to the command parameter definitions.
 *
 * In addition, AVRs are a little endian architecture, and so
 * multi-byte constants must be reversed to fascilitate setting
 * by way of casting.
 */
//@{
#define XBEE_AT_A1 0x3141 //!< Set/read end device association options. See \ref A1 for more details.
#define XBEE_AT_A2 0x3241 //!< Set/read coordinator association options. See \ref A2 for more details.
#define XBEE_AT_AC 0x4341 //!< Apply queued changes.
#define XBEE_AT_AI 0x4941 //!< Read association status. See \ref AI for more details.
#define XBEE_AT_AP 0x5041 //!< Set/read API enable.
#define XBEE_AT_BD 0x4442 //!< Set/read baud rate. See \ref BaudRates for more details.
#define XBEE_AT_CE 0x4543 //!< Set/read coordinator enable.
#define XBEE_AT_CH 0x4843 //!< Set/read channel. Possible channels: 0x0B-0x1A.
#define XBEE_AT_CN 0x4E43 //!< Exit AT command mode.
#define XBEE_AT_CT 0x5443 //!< Set/read command timeout (x100ms).
#define XBEE_AT_DA 0x4144 //!< Force disassociation.
#define XBEE_AT_DB 0x4244 //!< Read last received RSSI (-dBm).
#define XBEE_AT_ID 0x4449 //!< Set/read PAN ID.
#define XBEE_AT_FR 0x5246 //!< Firmware reset.
#define XBEE_AT_MY 0x594D //!< Set/read local module address.
#define XBEE_AT_SC 0x4353 //!< Set/read channel scan list.
#define XBEE_AT_SM 0x4D53 //!< Set/read sleep mode configuration.
#define XBEE_AT_SH 0x4853 //!< Read hardware address high word.
#define XBEE_AT_SL 0x4C53 //!< Read hardware address low word.
//@}

/**
 * \brief Cast queue location to 16-bit address.
 *
 * This macro casts a given queue position to a
 * 16-bit reference for AT command set convenience.
 * It is used only by the link layer and is not
 * intended for general use.
 *
 * \param queue is a pointer to the queue reference.
 * \param pos is a position offset to be added to the
 *        queue pointer.
 */
#define XBEE_AT_CMD(queue,pos) *((uint16_t*)(queue+pos))

/**
 * \defgroup BaudRates Baud Rates
 * \brief XBee baud rate options.
 *
 * XBee baud rate configuration commands as defined
 * by the XBee specification (pg. 37).
 *
 * See \ref XBEE_AT_BD for more details.
 */
//@{
#define XBEE_BAUD_1200   0x00 //!< 1200 bps.
#define XBEE_BAUD_2400   0x01 //!< 2400 bps.
#define XBEE_BAUD_4800   0x02 //!< 4800 bps.
#define XBEE_BAUD_9600   0x03 //!< 9600 bps.
#define XBEE_BAUD_19200  0x04 //!< 19200 bps.
#define XBEE_BAUD_38400  0x05 //!< 38400 bps.
#define XBEE_BAUD_57600  0x06 //!< 57600 bps.
#define XBEE_BAUD_115200 0x07 //!< 115200 bps.
//@}

/**
 * \defgroup A1 End Device Association Options
 * \brief End device association configuration options.
 *
 * End device association configuration options as
 * defined by the XBee specification (pg. 34).
 *
 * See \ref XBEE_AT_A1 for more details.
 */
//@{
#define XBEE_A1_REASSIGN_ID      0x01 //!< Automatically reassign PAN ID.
#define XBEE_A1_REASSIGN_CH      0x02 //!< Automatically reassign channel.
#define XBEE_A1_AUTO_ASSOCIATE   0x04 //!< Automatically associate with a coordinator.
#define XBEE_A1_POLL_ON_PIN_WAKE 0x08 //!< Poll coordinator on pin wakeup.
//@}

/**
 * \defgroup A2 Coordinator Association Options
 * \brief Coordinator association configuration options.
 *
 * Coordinator association configuration options as
 * defined by the XBee specification (pg. 34).
 *
 * See \ref XBEE_AT_A2 for more details.
 */
//@{
#define XBEE_A2_REASSIGN_ID       0x01 //!< Automatically reassign PAN ID.
#define XBEE_A2_REASSIGN_CH       0x02 //!< Automatically reassign channel.
#define XBEE_A2_ALLOW_ASSOCIATION 0x04 //!< Allow end devices to associate.
//@}

/**
 * \defgroup AI Association Indicators
 * \brief Device association indication.
 *
 * Device association indication values as
 * defined by the XBee specification (pg. 35).
 *
 * See \ref XBEE_AT_AI for more details.
 */
//@{
#define XBEE_AI_SUCCESS                      0x00 //!< Coordinator successfully started / end device association complete.
#define XBEE_AI_ACTIVE_SCAN_TIMEOUT          0x01 //!< Active PAN scan timeout.
#define XBEE_AI_ACTIVE_SCAN_NONE             0x02 //!< Active scan found no PANs.
#define XBEE_AI_ACTIVE_SCAN_RESTRICTED       0x03 //!< Active scan found PAN, but association is not allowed.
#define XBEE_AI_ACTIVE_SCAN_NO_BEACONS       0x04 //!< Active scan found PAN, but network not configured to support beacons.
#define XBEE_AI_ACTIVE_SCAN_ID_MISMATCH      0x05 //!< Active scan found PAN, but PAN ID does not match configuration.
#define XBEE_AI_ACTIVE_SCAN_CHANNEL_MISMATCH 0x06 //!< Active scan found PAN, but channel does not match configuration.
#define XBEE_AI_ENERGY_SCAN_TIMEOUT          0x07 //!< Energy scan timeout.
#define XBEE_AI_COORDINATOR_START_FAILED     0x08 //!< Coordinator start request failed.
#define XBEE_AI_COORDINATOR_INVALID_PARAM    0x09 //!< Coordinator start request failed due to invalid parameter.
#define XBEE_AI_REALIGNMENT_IN_PROGRESS      0x0A //!< Coordinator realignment in progress.
#define XBEE_AI_ASSOCIATION_REQ_NOT_SENT     0x0B //!< Association request not sent.
#define XBEE_AI_ASSOCIATION_REQ_TIMEOUT      0x0C //!< Association request timed out - no reply was received.
#define XBEE_AI_ASSOCIATION_REQ_INVALID      0x0D //!< Association request has an invalid parameter.
#define XBEE_AI_ASSOCIATION_REQ_CCA_FAILURE  0x0E //!< Association request channel access failure (CCA failure).
#define XBEE_AI_ASSOCIATION_REQ_ACK_LOST     0x0F //!< Coordinator did not send an ACK after association request was sent.
#define XBEE_AI_ASSOCIATION_REQ_REPLY_LOST   0x10 //!< Coordinator did not reply to the association request, but an ACK was received.
#define XBEE_AI_SYNC_LOST                    0x12 //!< Lost synchronization with beaconing coordinator.
#define XBEE_AI_DISASSOCIATED                0x13 //!< No longer associated to a coordinator.
#define XBEE_AI_ASSOCIATING                  0xFF //!< RF module is attempting to associate.
//@}

/**
 * \defgroup LinkTXPos Link TX Frame Structure
 * \brief Component positions for 64-bit link TX frames.
 *
 * Transmit frames contain an API and frame identifier
 * (the frame ID 0 disables response), a destination address,
 * an RSSI reading, and an options field.
 */
//@{
#define LINK_TX_POS_API_ID  0  //!< API ID.
#define LINK_TX_POS_FRAME   1  //!< Frame ID. Set to 0 to disable TX response.
#define LINK_TX_POS_DEST    2  //!< 64-bit destination address.
#define LINK_TX_POS_OPTIONS 10 //!< TX options bitmask. See \ref LinkTXOptions for more details.
#define LINK_TX_POS_DATA    11 //!< Frame payload.
//@}

/**
 * \defgroup LinkRXPos Link RX Frame Structure
 * \brief Component positions for 64-bit link RX frames.
 *
 * Receive frames contain a source address, an RSSI reading,
 * and an options field.
 */
//@{
#define LINK_RX_POS_SOURCE  0  //!< 64-bit source address.
#define LINK_RX_POS_RSSI    8  //!< RSSI value in -dBm.
#define LINK_RX_POS_OPTIONS 9  //!< RX options bitmask. See \ref LinkRXOptions for more details.
#define LINK_RX_POS_DATA    10 //!< Frame payload.
//@}

/**
 * \defgroup LinkStatusPos Link Status Frame Structure
 * \brief Component positions for link status frames.
 *
 * Status frames are sent by the XBee module on network
 * configuration changes. See \ref LinkStatus for more details.
 */
//@{
#define LINK_STATUS_POS_DATA 0 //!< Frame payload.
//@}

/**
 * \defgroup LinkATPos Link AT Response Frame Structure
 * \brief Component positions for AT command response frames.
 *
 * AT command responses contain an 8-bit ID field, a 16-bit
 * command type, an 8-bit status value, and a payload.
 */
//@{
#define LINK_AT_POS_ID     0 //!< Command ID.
#define LINK_AT_POS_CMD    1 //!< Command type.
#define LINK_AT_POS_STATUS 3 //!< Command status (0=OK, 1=Error).
#define LINK_AT_POS_DATA   4 //!< Frame payload.
//@}

/**
 * \defgroup LinkTXOptions Link TX Frame Options
 * \brief Options for a link TX frame.
 *
 * Transmit frames may be sent with no receiver
 * acknowledgement expectation. Additionally, they
 * can be broadcast to any PAN ID.
 */
//@{
#define LINK_TX_OPTION_NO_ACK    0x01 //!< Disable receiver frame ACK.
#define LINK_TX_OPTION_BROADCAST 0x04 //!< Send with broadcast PAN ID.
//@}

/**
 * \brief Queue index increment macro.
 *
 * This macro increments the specified circular buffer
 * index using the supplied index mask. It is implemented
 * using the \ref LINK_INDEX_OFFSET macro and is for
 * convenience only.
 *
 * \param index is the index to be incremented.
 * \param mask is the index mask to be used.
 */
#define LINK_INDEX_INC(index,mask) index=LINK_INDEX_OFFSET(index,1,mask);

/**
 * \brief Queue index offset macro.
 *
 * This macro offsets the specified circular buffer
 * index by a given amount, and wraps it using the
 * supplied index mask.
 *
 * \param index is the index to be incremented.
 * \param offset is the offset value to be added to the index.
 * \param mask is the index mask to be used.
 */
#define LINK_INDEX_OFFSET(index,offset,mask) ((index+offset) & mask)

/**
 * \defgroup USART USART Helper Macros
 * \brief Helper macros for generic interfaces.
 *
 * These macros allow for a more generic interface
 * to the USART(s).
 */
//@{

/**
 * \brief USART baud rate helper macro.
 *
 * This macro generates a value for the UBRR register
 * given a desired baud rate and the specified CPU
 * clock frequency using the following formula:
 *
 *    UBRR = fosc/16/baud-1
 *
 * \param baud is the desired baud rate.
 */
#define BAUD(baud) (((F_CPU>>4)/baud)-1)

/**
 * \brief Check if USART is free for data.
 *
 * This macro returns whether or not the USART is
 * available for the specified interface. It is used
 * only by the link layer and is not intended for
 * general use.
 *
 * \param interface is the unique identifier for the interface.
 */
#if AVR_NUM_INTERFACES==2
#define LINK_USART_READY(interface) !(UCSR ## interface ## B & _BV(UDRIE ## interface))
#else
#define LINK_USART_READY(interface) !(UCSRB & _BV(UDRIE))
#endif

/**
 * \brief Reset specified module.
 *
 * This function resets the XBee module on the
 * specified interface, and waits for the reset
 * to finish.
 *
 * \param interface is the unique identifier for the interface.
 */
void Link_ResetModule(unsigned char interface);

/**
 * \brief Send specified AT command.
 *
 * Send the given AT command to the XBee module,
 * and wait for a successful response confirming
 * the command if so desired.
 *
 * \param interface is the unique identifier for the interface.
 * \param data is a pointer to the command to be sent.
 * \param length is the length of the command to be sent.
 */
void Link_SendATCommand(uint8_t interface, uint8_t* data, uint8_t length, uint8_t wait);

/**
 * \brief Send XBee API frame.
 *
 * Format packet into an XBee API frame and send it on the interface.
 *
 * \param interface is the unique identifier for the interface.
 * \param data is a pointer to the packet to be sent.
 * \param length is the length of the packet to be sent.
 *
 * \return Returns 1 if succeeded, 0 otherwise.
 */
uint8_t Link_SendAPIFrame(unsigned char interface, unsigned char* data, unsigned char length);

/**
 * \brief Determine if a frame is pending.
 *
 * Return whether or not a frame is pending in the
 * interface receive queue.
 *
 * \param interface is the unique identifier for the interface.
 *
 * \return Returns 1 if a frame is pending, 0 otherwise.
 */
inline uint8_t Link_FramePending(uint8_t interface);

/**
 * \brief Determine length of next frame.
 *
 * Determine the total length (in bytes) of the
 * frame currently at the head of the interface
 * receive queue.
 *
 * \param interface is the unique identifier for the interface.
 *
 * \return The length of the next frame.
 */
inline uint8_t Link_FrameLength(uint8_t interface);

/**
 * \brief Determine payload length of the next frame.
 *
 * Determine the total payload length (in bytes) from
 * the frame currently at the head of the interface
 * receive queue.
 *
 * \param interface is the unique identifier for the interface.
 *
 * \return The length of the payload of the next frame.
 */
inline uint8_t Link_PayloadLength(uint8_t interface);

//@}

#endif
