/************************************************************
* This file is part of the XDN Router project.
*
* XDN Router is a packet routing device to be used with
* a local XDN (XBee Distributed Network) configuration.
* It supports both core and edge routing modes.
* Copyright (C) 2008 Adam Shapiro, Cornell University
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
************************************************************/
/** \file
 * \brief Router definitions.
 *
 * This file defines router modes and
 * various configuration parameters.
 */

#ifndef __router_h__
#define __router_h__

#include <avr/io.h>
#include "link_interfaces.h"

/**
 * \brief Manual router disable.
 *
 * This macro allows a device to override
 * the default router definition in order
 * to use a second serial interface for
 * other purposes.
 *
 * When set, the first interface (USART0)
 * is used for XDN communication. The second
 * interface (USART1) is available for
 * other use.
 */
#define ROUTER_DISABLE 1

/**
 * \brief Router enable macro.
 *
 * A device can only be classified as
 * a router if it has at least two
 * dedicated network interfaces.
 */
#define IS_ROUTER (AVR_NUM_INTERFACES==2 && !ROUTER_DISABLE)	// = 0

/**
 * \brief Number of link layer interfaces.
 *
 * This macro defines the number of interfaces
 * available to the link layer, depending on
 * whether or not the device is configured as
 * a router.
 */
#define LINK_NUM_INTERFACES (IS_ROUTER ? 2 : 1)					// = 1

/**
 * \defgroup IgnoreRadios Radio Ignore Enables
 * \brief Radio ignore enables.
 *
 * These macros allow the link layer to ignore
 * the specified interface when initializing.
 * This might be desirable if a radio is not
 * attached, in which case command responses will
 * not be received.
 */
//@{
#define LINK_IGNORE_RADIO_0 0
#define LINK_IGNORE_RADIO_1 0
//@}

/**
 * \defgroup DebugLEDs Debug LED Macros
 * \brief Debug LED macros.
 *
 * These macros enable a number of debugging
 * LEDs on the specified port. To use, set
 * the value to the desired port letter.
 * For example, to debug using port B specify
 * the following:
 *
 * - ENABLE_DEBUG_LEDS = 1
 * - DEBUG_PORT_LETTER = B
 *
 * The following macros can be used to control
 * the debug LEDs:
 *
 * - \ref DEBUG_ON
 * - \ref DEBUG_OFF
 * - \ref DEBUG_TOGGLE
 *
 * For example, to toggle LED 7 do:
 *
 * \ref DEBUG_TOGGLE(7)
 */
//@{
#define ENABLE_DEBUG_LEDS 0
#define DEBUG_PORT_LETTER B
#define DEBUG_CAT(a,b) a ## b
#define DEBUG_DDR_EXP(x) DEBUG_CAT(DDR,x)
#define DEBUG_DDR DEBUG_DDR_EXP(DEBUG_PORT_LETTER)
#define DEBUG_PORT_EXP(x) DEBUG_CAT(PORT,x)
#define DEBUG_PORT DEBUG_PORT_EXP(DEBUG_PORT_LETTER)

#if ENABLE_DEBUG_LEDS
#define DEBUG_ON(x) DEBUG_ON_BLOCK(_BV(x))
#define DEBUG_OFF(x) DEBUG_OFF_BLOCK(_BV(x))
#define DEBUG_TOGGLE(x) DEBUG_TOGGLE_BLOCK(_BV(x))
#define DEBUG_ON_BLOCK(x) DEBUG_PORT&=~(x)
#define DEBUG_OFF_BLOCK(x) DEBUG_PORT|=(x)
#define DEBUG_TOGGLE_BLOCK(x) DEBUG_PORT^=(x)
#define DEBUG_IS_ON(x) (DEBUG_PORT&_BV(x))
#else
#define DEBUG_ON(x)
#define DEBUG_OFF(x)
#define DEBUG_TOGGLE(x)
#define DEBUG_ON_BLOCK(x)
#define DEBUG_OFF_BLOCK(x)
#define DEBUG_TOGGLE_BLOCK(x)
#define DEBUG_IS_ON(x) 0
#endif
//@}

/**
 * \brief Enable debug LEDs.
 *
 * This macro enables a number of debugging
 * LEDs on the specified port. To use, set
 * the value to the desired port macro.
 * For example, to use port B do:
 *
 * #define DEBUG_LEDS PORTB
 */
#define DEBUG_LEDS PORTB

/**
 * \brief Force link layer broadcasting.
 *
 * This flag forces the link layer to broadcast
 * frames, regardless of destination address.
 * This is for debugging purposes only, and
 * should be disabled during normal operation.
 */
#define FORCE_BROADCAST 0

/**
 * \brief Router mode.
 *
 * Routers may be configured as either
 * core or edge routers. Core routers
 * implement the Route Control Protocol
 * and adopt an upstream address after
 * building a routing table. Edge routers,
 * on the other hand, forward packets to
 * their local subnet or gateway as necessary,
 * and acquire their an upstream address
 * by way of the Dynamic Addressing Protocol.
 *
 * See \ref RouterModes.
 */
#define ROUTER_MODE ROUTER_MODE_CORE

/**
 * \defgroup RouterModes Router Modes
 * \brief Router mode definitions.
 */
//@{
#define ROUTER_MODE_CORE 0 //!< Core routing mode.
#define ROUTER_MODE_EDGE 1 //!< Edge routing mode.
//@}

/**
 * \brief Core router determination.
 *
 * This macro evaluates to true if the
 * current device is a router, configured
 * in CORE mode.
 *
 * See \ref IS_ROUTER and ROUTER_MODE
 * for more details.
 */
#define IS_CORE_ROUTER (IS_ROUTER && ROUTER_MODE==ROUTER_MODE_CORE)		// = 0

/**
 * \brief Edge router determination.
 *
 * This macro evaluates to true if the
 * current device is a router, configured
 * in EDGE mode.
 *
 * See \ref IS_ROUTER and ROUTER_MODE
 * for more details.
 */
#define IS_EDGE_ROUTER (IS_ROUTER && ROUTER_MODE==ROUTER_MODE_EDGE)		// = 0

/**
 * \brief Edge device determination.
 *
 * This macro evaluates to true if the
 * current device is an edge device -
 * either a single-interface device or
 * a router configured in EDGE mode.
 *
 * See \ref IS_ROUTER and ROUTER_MODE
 * for more details.
 */
#define IS_EDGE_DEVICE (!IS_ROUTER || ROUTER_MODE==ROUTER_MODE_EDGE)	// = 0

#endif
