/************************************************************
* This file is part of the XDN Router project.
*
* XDN Router is a packet routing device to be used with
* a local XDN (XBee Distributed Network) configuration.
* It supports both core and edge routing modes.
* Copyright (C) 2008 Adam Shapiro, Cornell University
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
************************************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "xdn/link.h"

#include <stdio.h>
#include "uart.h"
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

// bit handling macros for i/o registers
#define READ(U, N) ((U) >> (N) & 1u)
#define SET(U, N) ((void)((U) |= 1u << (N)))
#define CLR(U, N) ((void)((U) &= ~(1u << (N))))
#define FLIP(U, N) ((void)((U) ^= 1u << (N)))

uint8_t Ain, junk;
uint8_t x_hi_data[15], y_hi_data[15], z_hi_data[15], savedTime[15];
uint8_t prevXhi, prevYhi, prevZhi;
long sleepCount;
char startup, sleeping;
char led, dataCounter, newData;

uint8_t data[60], length;
uint8_t data0[15],data1[15],data2[15];
int i,j;
MACAddress TranAddress;

volatile int msTimer, sysTime;
volatile char dataFlag;

void InitializeTimers(void);
void Link_UpdateTimers(void);
void get_Accelerometer(void);

#define CTRL_REGB  	0x0D  	//0b00001101;
#define CTRL_REGC  	0x0C  	//0b00001100;
#define XOUT_HI  	0x80	  	//0b10000000;
#define	XOUT_LO  	0x81		//0b10000001;
#define YOUT_HI  	0x82		//0b10000010;
#define YOUT_LO  	0x83		//0b10000011;
#define ZOUT_HI  	0x84
#define ZOUT_LO  	0x85

#define REGB_DEFAULT  0x40 //0b01000000; //0b11000000;
#define REGC_DEFAULT  0xF2 //0b11100011; //0b11110010;
//*****
//CTRL_REGB
//bit 7 CLKhld = 0	SCL not held low during A/D conversion
//bit 6 ENABLE = 1	Normal operation
//bit 5 ST = 0		no self-test
//bit 4 0
//bit 3 0
//bit 2 MOTlen = 0	Normal operating mode
//bit 1 0
//bit 0 0

//CTRL_REGC
//bit 7 LP2 = 1		50Hz filter
//bit 6 LP1 = 1
//bit 5 LP0 = 1
//bit 4 MOTLev = 1	wake-up threshold +/- 3g
//bit 3 MOTLat = 0	data sheet diagrams (set by MOTlen)
//bit 2 0
//bit 1 FS1 = 0		+/- 6g
//bit 0 FS0 = 1
//*****
	
int main(void)
{
    
    InitializeTimers();

	DDRD = 0xff;
	PORTD = 0x00;
    
	uart_init();
  	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"Starting timers...\n\r");

	//leds
    DDRC=0xff;
    PORTC=0xff;

	//****************************
	//transmitter
	//**************************
	dataCounter = 0;
	dataFlag = 0;
	newData = 0;

//	data0 = 0x00;
//	data1 = 0x00;
//	data2 = 0x00;

	msTimer = 0;

	//****************************
	//SPI
	//**************************
	//set up i/o data direction
	//B.0, B.1 unused
	//B.2 = MOT, input
	//B.3 = Aux_in, input
    //DDRB.4 = 1; //output chip select for DAC (notSYNC)
    //DDRB.5 = 1; //output MOSI to ADC
    //DDRB.6 = 0; //input MISO from ADC
    //DDRB.7 = 1; //output SCLK   
	DDRB = (1<<DDB4) | (1<<DDB5) | (1<<DDB7) ; 
	SET(PORTB,4);

	//set up SPI
    //bit 7 SPIE=0 no ISR
    //bit 6 SPE=1 enable spi
    //bit 5 DORD=0 msb first 
    //bit 4 MSTR=1 Mega32 is spi master
    //bit 3 CPLO=1 clock polarity
    //bit 2 CPHA=1 clock phase
    //bit 1,0 rate sel=10 along with SPRC=1 sets clk to f/16 = 1MHz  (SPI2X = SPR1 = 0, SPR0 = 1)
    SPCR = (1<<SPE) | (1<<MSTR) | (0<<CPOL) | (0<<CPHA) | (1<<SPR0);                                                                               
    SPSR = (0<<SPI2X) ; //SPSR = 1;
	sei();

	sleepCount = 0;
	sleeping = 0;
	startup = 1;
	prevXhi = 0;
	prevYhi = 0;
	prevZhi = 0;

	//****************************
	//Init accelerometer
	CLR(PORTB,4) ;   //chip select low begins conversion 
	//set acceleration parameters               
    SPDR = CTRL_REGB; 		
	//wait until SPIF is set (SPI done)
    while (!(SPSR & (1<<SPIF))) ;
	junk = SPDR ; // get back junk
	//Now send junk to get back ADC value
    SPDR = REGB_DEFAULT ;  
	//wait until SPIF is set (SPI done)
    while (!(SPSR & (1<<SPIF))) ;
	junk = SPDR ;
	SET(PORTB,4);	//chip select high

	_delay_us(1);	//wait at least 130 ns before next conversion
		

	CLR(PORTB,4);
	SPDR = CTRL_REGC;     
	//wait until SPIF is set (SPI done)
    while (!(SPSR & (1<<SPIF))) ;
	junk = SPDR ; // get back junk
	//Now send data to store
    SPDR = REGC_DEFAULT ;  
	//wait until SPIF is set (SPI done)
    while (!(SPSR & (1<<SPIF))) ;
	junk = SPDR;
	//SYNC high ends load
    SET(PORTB,4) ;  //chip select high ends conversion
		
	_delay_us(1);


	//*************************
	//Init TX
	//DEVICE_TYPE=1 //transmitter
	Link_Init(0);

	//0x0013A200403C0617	receiver address
	//0x0013A200403C0619	transmitter address
	TranAddress.addr64 = 0x0013A200403C0617;	//Receiving address
    
    while(1)
    {
		if(dataFlag == 1 && dataCounter < 15) {
			//get sample from the accelerometer every 1 kHz as timed by the ISR
			dataFlag = 0;
			get_Accelerometer();
		}

		//if the accelerometer has not changed significantly for 5 minutes
		if(sleepCount >= 300000)
		{
			SET(PORTD,4);	//assert the sleep pin on the XBee
			sleeping = 1;	//set flag so we know that it is sleeping
		}

		//Accelerometer has now changed significantly, so start back up if sleeping
		if(sleeping && startup)
		{
			CLR(PORTD,4);	//deassert the sleep pin
			sleepCount = 0;	//reset all flags
			startup = 0;
			sleeping = 0;
		}

		//send 15 data samples in a packet
		if(dataCounter == 15)
		{		
			i = 0;
			j = 0;
			while(j < 15) {						//format packet
				data[i] = x_hi_data[j]; i++;
				data[i] = y_hi_data[j]; i++;
				data[i] = z_hi_data[j]; i++;
				data[i] = savedTime[j]; i++;
				j++;
			}
		
			length = 60;						//length of payload
			
			Link_Execute();						//check XBee link
			if(sleeping==0)						//if transmitter is not asleep, send the packet
				led = Link_SendFrame(0, &TranAddress, data, length); 

			dataCounter = 0;					//reset for the next 15 data points
		}		
    }

    return 0;
}

//********************************************
//Accelerometer
//SPI interface
//********************************************
void get_Accelerometer(void)
{
	if(newData == 0) {
		CLR(PORTB,4);				//set chip select to low to begin SPI
	 	SPDR = XOUT_HI;				//send address of desired register
		while(!(SPSR & (1<<SPIF)));
		junk = SPDR;
		SPDR = junk;
		while(!(SPSR & (1<<SPIF)));
		//hold sclk low for 200us to allow time for A/D conversion
		_delay_us(200);
		x_hi_data[dataCounter] = SPDR;	//store the returned data into the x_hi_data array
		SET(PORTB,4);

		_delay_us(1);					//delay in between samples as required by accelerometer
		
		CLR(PORTB,4);
	 	SPDR = YOUT_HI;
		while(!(SPSR & (1<<SPIF)));
		junk = SPDR;
		SPDR = junk;
		while(!(SPSR & (1<<SPIF)));
		//hold sclk low for 200us
		_delay_us(200);
		y_hi_data[dataCounter] = SPDR;
		SET(PORTB,4);

		_delay_us(1);

		CLR(PORTB,4);
	 	SPDR = ZOUT_HI;
		while(!(SPSR & (1<<SPIF)));
		junk = SPDR;
		SPDR = junk;
		while(!(SPSR & (1<<SPIF)));
		//hold sclk low for 200us
		_delay_us(200);
		z_hi_data[dataCounter] = SPDR;
		SET(PORTB,4);

		savedTime[dataCounter] = msTimer;		//get the time of the data aquisition so we can plot data vs. time

		_delay_us(1);

		//check to see if this set of data is the same as the previous set of data
		//in order to count how many have been the same to send it to sleep if 
		// the accelerometer has not been moved significantly for a long period of time
		if((abs(x_hi_data[dataCounter] - prevXhi) < 20)&&(abs(y_hi_data[dataCounter] - prevYhi) < 20)&&(abs(z_hi_data[dataCounter] - prevZhi) < 20)){
			if(!sleeping)			//if it is already sleeping, then don't increase count.
				sleepCount++;
		}
		else 		//the accelerometer has moved significantly
		{		
			if(sleeping == 1)			//if it is already sleeping, set the start-up flag so that the transmitter will wake up
				startup = 1;
			else						//if it is already tranmitting, then reset sleepCount to start counting from 0 and make sure startup does nothing
			{
				startup = 0;
				sleepCount=0;
			}
		}

		//store the current data as previous data for the next round of SPI data aquisition
		prevXhi = x_hi_data[dataCounter];
		prevYhi = y_hi_data[dataCounter];
		prevZhi = z_hi_data[dataCounter];

		//increase dataCounter
		dataCounter++;
	}
	//newData = newData ^ 0x01;
}

void InitializeTimers(void)
{
    //Timer 0 is set to compare/match, interrupting
    //with a 1ms interval. With a 16MHz clock using
    //a 1/64 clock divider, a compare value of 250
    //is required to produce an exact 1ms interrupt,
    //with the accuracy of the crystal itself.
    TCCR0A=_BV(WGM01);//Clear on compare match.
    TCCR0B=_BV(CS01) | _BV(CS00);//fclk/64.
    OCR0A=250;
    TIMSK0=_BV(OCIE0A);
}

ISR(TIMER0_COMPA_vect)
{
    Link_UpdateTimers();
	msTimer++;
	dataFlag = 1;
}
